/**
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URL;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.component.ComponentContext;
import javax.jbi.component.ComponentLifeCycle;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.management.ObjectName;
import javax.xml.namespace.QName;

import org.ow2.easywsdl.wsdl.api.Endpoint;
import org.ow2.easywsdl.wsdl.api.WSDLException;
import org.ow2.petals.component.framework.api.Constants;
import org.ow2.petals.component.framework.api.Interceptor;
import org.ow2.petals.component.framework.api.PetalsComponent;
import org.ow2.petals.component.framework.api.Wsdl;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.notification.NotificationBrokerController;
import org.ow2.petals.component.framework.api.notification.NotificationController;
import org.ow2.petals.component.framework.jbidescriptor.CDKJBIDescriptorException;
import org.ow2.petals.component.framework.jbidescriptor.JBIDescriptorBuilder;
import org.ow2.petals.component.framework.jbidescriptor.generated.ComponentInterceptor;
import org.ow2.petals.component.framework.jbidescriptor.generated.Jbi;
import org.ow2.petals.component.framework.mbean.InterceptorManager;
import org.ow2.petals.component.framework.mbean.MBeanHelper;
import org.ow2.petals.component.framework.mbean.RuntimeConfigurationNotifier;
import org.ow2.petals.component.framework.notification.NotificationService;
import org.ow2.petals.component.framework.process.JBIAcceptorManager;
import org.ow2.petals.component.framework.process.JBIProcessorManager;
import org.ow2.petals.component.framework.process.async.AsyncMessageManager;
import org.ow2.petals.component.framework.su.AbstractServiceUnitManager;
import org.ow2.petals.component.framework.util.ServiceEndpointKey;
import org.ow2.petals.component.framework.util.StringHelper;
import org.ow2.petals.component.framework.util.UtilFactory;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;

import com.ebmwebsourcing.wsstar.notification.definition.utils.WSNotificationException;

/**
 * An abstract class that can be used to easily create a new JBI Component.<br />
 * It implements the Component interface.<br />
 * It allow to manage directly endpoints.<br />
 * To do this, simply put into the resources of the project a valid wsdl named
 * 'component.wsdl'
 * 
 * @author ofabre - EBMWebsourcing
 * @author fgardes - EBMWebsourcing
 * @since 1.0
 * 
 */
public abstract class AbstractComponent implements PetalsComponent {

    /**
     * The Component's Context
     */
    protected ComponentContext context;

    /**
     * Notifications manager
     */
    private org.ow2.petals.component.framework.api.notification.NotificationService notificationService;

    /**
     * The message exchange acceptor manager
     */
    private JBIAcceptorManager acceptorManager;

    /**
     * The asynchronous message handler
     */
    private AsyncMessageManager asyncManager;

    /**
     * The component's delivery channel
     */
    private DeliveryChannel channel;

    /**
     * The component specific configuration in extensions form.
     */
    private ConfigurationExtensions componentExtensions;

    /**
     * The list of endpoints, managed directly by the component
     */
    private Map<ServiceEndpointKey, ServiceEndpoint> componentServiceEndpoints;

    /**
     * Manage the interceptors at runtime, exposed as MBean
     */
    private InterceptorManager interceptorManager;

    /**
     * All the component interceptors. Key is the name of the interceptor
     */
    private Map<String, Interceptor> interceptors;

    /**
     * The component configuration
     */
    private Jbi jbiComponentConfiguration;

    /**
     * The component's logger
     */
    private Logger logger;

    /**
     * The description and the document of the native WSDL managed internally by
     * the component
     */
    private Wsdl nativeWsdl;

    /**
     * The processor manager in charge of handling JBI Message Exchanges
     */
    private JBIProcessorManager processorManager;

    /**
     * Manage runtime configuration notifications, exposed as MBean
     */
    private RuntimeConfigurationNotifier runtimeConfigurationNotifier;

    /**
     * The component's service unit manager
     */
    private AbstractServiceUnitManager suManager;

    public AsyncMessageManager getAsyncManager() {
        return this.asyncManager;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.Component#getChannel()
     */
    public DeliveryChannel getChannel() {
        return this.channel;
    }

    public org.ow2.petals.component.framework.jbidescriptor.generated.Component getComponentConfiguration() {
        return this.jbiComponentConfiguration.getComponent();
    }

    public ConfigurationExtensions getComponentExtensions() {
        return this.componentExtensions;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.Component#getContext()
     */
    public ComponentContext getContext() {
        return this.context;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.Component#getExtensionMBeanName()
     */
    public ObjectName getExtensionMBeanName() {
        return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.Component#getInterceptor(java.lang
     * .String)
     */
    public Interceptor getInterceptor(final String name) {
        Interceptor interceptor = null;

        if (this.interceptors != null) {
            interceptor = this.interceptors.get(name);
        }
        return interceptor;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.Component#getInterceptors()
     */
    public Map<String, Interceptor> getInterceptors() {
        return this.interceptors;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.Component#getLifeCycle()
     */
    public ComponentLifeCycle getLifeCycle() {
        return this;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.Component#getLogger()
     */
    public Logger getLogger() {
        return this.logger;
    }

    /**
     * Get the nativeWsdl
     * 
     * @return the nativeWsdl
     */
    public Wsdl getNativeWsdl() {
        return this.nativeWsdl;
    }

    /**
     * @return the {@link NotificationBrokerController}, can be null...
     * @deprecated
     * @deprecated use the {@link NotificationService} instead.
     */
    @Deprecated
    public NotificationBrokerController getNotificationBrokerController() {
        return this.notificationService.getNotificationBrokerController();
    }

    /**
     * @return the
     *         {@link org.ow2.petals.component.framework.notification.NotificationContext}
     *         , can be null
     * @deprecated use the {@link NotificationService} instead.
     */
    @Deprecated
    public NotificationController getNotificationController() {
        return this.notificationService.getNotificationController();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.Component#getServiceDescription
     * (javax.jbi.servicedesc.ServiceEndpoint)
     */
    public Document getServiceDescription(final ServiceEndpoint endpoint) {
        if (endpoint == null) {
            return null;
        }
        try {
            if ((this.componentServiceEndpoints != null)
                    && !this.componentServiceEndpoints.isEmpty()
                    && (this.componentServiceEndpoints.get(new ServiceEndpointKey(endpoint
                            .getServiceName(), endpoint.getEndpointName())) != null)) {

                if ((this.nativeWsdl != null)
                        && UtilFactory.getWSDLUtil().isDescriptionContaining(
                                this.nativeWsdl.getDescription(), endpoint)) {
                    return this.nativeWsdl.getDocument();
                    // TODO remove the controller without managers
                } else if (this.getNotificationBrokerController() != null) {
                    if ((this.getNotificationBrokerController().getNotificationProducer() != null)
                            && (this.getNotificationBrokerController().getNotificationProducer()
                                    .getWsdl() != null)
                            && (this.getNotificationBrokerController().getNotificationProducer()
                                    .getWsdl().getDescription() != null)
                            && UtilFactory.getWSDLUtil().isDescriptionContaining(
                                    this.getNotificationBrokerController()
                                            .getNotificationProducer()
                                            .getWsdl().getDescription(), endpoint)) {
                        return this.getNotificationBrokerController().getNotificationProducer()
                                .getWsdl().getDocument();
                    } else if ((this.getNotificationBrokerController().getNotificationConsumer() != null)
                            && (this.getNotificationBrokerController().getNotificationConsumer()
                                    .getWsdl() != null)
                            && (this.getNotificationBrokerController().getNotificationConsumer()
                                    .getWsdl().getDescription() != null)
                            && UtilFactory.getWSDLUtil().isDescriptionContaining(
                                    this.getNotificationBrokerController()
                                            .getNotificationConsumer()
                                            .getWsdl().getDescription(), endpoint)) {
                        return this.getNotificationBrokerController().getNotificationConsumer()
                                .getWsdl().getDocument();
                    }
                } else if (this.getNotificationController() != null) {
                    if ((this.getNotificationController().getProducerWsdl() != null)
                            && (this.getNotificationController().getProducerWsdl().getDescription() != null)
                            && UtilFactory.getWSDLUtil().isDescriptionContaining(
                                    this.getNotificationController().getProducerWsdl()
                                            .getDescription(),
                                    endpoint)) {
                        return this.getNotificationController().getProducerWsdl().getDocument();
                    } else if ((this.getNotificationController().getConsumerWsdl() != null)
                            && (this.getNotificationController().getConsumerWsdl().getDescription() != null)
                            && UtilFactory.getWSDLUtil().isDescriptionContaining(
                                    this.getNotificationController().getConsumerWsdl()
                                            .getDescription(),
                                    endpoint)) {
                        return this.getNotificationController().getConsumerWsdl().getDocument();
                    }
                }
            }
        } catch (final WSDLException e) {
            this.logger.warning("Unable to found the endpoint managed by the component: "
                    + e.getMessage() + ". Search it into the Service Unit Manager...");
        }
        return this.suManager.getServiceDescription(endpoint);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.Component#getServiceUnitManager()
     */
    public javax.jbi.component.ServiceUnitManager getServiceUnitManager() {
        return this.suManager;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.Component#init(javax.jbi.component
     * .ComponentContext)
     */
    public final void init(final ComponentContext compContext) throws JBIException {
        this.context = compContext;
        this.logger = this.context.getLogger("", this.getResourceBundleName());
        this.logger.info("Init component...");
        this.channel = this.context.getDeliveryChannel();
        // Get the component configuration
        final File jbiXmlFile = new File(compContext.getInstallRoot() + File.separator
                + Constants.Component.META_INF + File.separator + Constants.Component.JBI_XML);
        try {
            // CDK configuration
            final FileInputStream jbiInputStream = new FileInputStream(jbiXmlFile);
            try {
                this.jbiComponentConfiguration = JBIDescriptorBuilder
                        .buildJavaJBIDescriptor(jbiInputStream);
            } finally {
                try {
                    jbiInputStream.close();
                } catch (final IOException e) {
                    // do nothing
                }
            }

            // Component specific configuration
            if ((this.jbiComponentConfiguration.getComponent().getPropertiesFile() != null)
                    && !StringHelper.isNullOrEmpty(this.jbiComponentConfiguration.getComponent()
                            .getPropertiesFile().getValue())) {
                this.componentExtensions = new ConfigurationExtensions(
                        this.jbiComponentConfiguration.getComponent().getAny(),
                        this.jbiComponentConfiguration.getComponent().getPropertiesFile()
                                .getValue());
            } else {
                this.componentExtensions = new ConfigurationExtensions(
                        this.jbiComponentConfiguration.getComponent().getAny());
            }
            this.runtimeConfigurationNotifier = new RuntimeConfigurationNotifier(
                    this.jbiComponentConfiguration, jbiXmlFile, this.logger);
            this.interceptors = this.loadInterceptors();
            this.interceptorManager = new InterceptorManager(this, this.jbiComponentConfiguration,
                    jbiXmlFile);
            // Necessary resources must be instantiated only once during the
            // life-cycle of the component
            if (this.suManager == null) {
                this.suManager = this.createServiceUnitManager();
            }

            final URL componentWsdlUrl = AbstractComponent.class.getResource("/"
                    + Constants.Component.COMPONENT_WSDL);
            if (componentWsdlUrl != null) {
                // initialize the description of the specific wsdl
                this.nativeWsdl = new ComponentWsdl(UtilFactory.getWSDLUtil()
                        .createWsdlDescription(componentWsdlUrl, true));
            }

            // initialize the notification service
            this.notificationService = new NotificationService(this);
            this.notificationService.init();
            // initialize the service endpoint list managed by the component
            this.componentServiceEndpoints = new ConcurrentHashMap<ServiceEndpointKey, ServiceEndpoint>();
        } catch (final CDKJBIDescriptorException e) {
            throw new JBIException(e);
        } catch (final FileNotFoundException e) {
            throw new JBIException(e);
        } catch (final PEtALSCDKException e) {
            throw new JBIException(e);
        } catch (final WSDLException e) {
            throw new JBIException(e);
        } catch (final WSNotificationException e) {
            throw new JBIException(e);
        }
        this.logger.fine("Calling initialization method on component implementation...");
        this.doInit();
        this.logger.fine("Specific initialization done.");
        // Since the worker manager instantiates the JBIListeners, we must
        // initialize it after the component specific initialization. This is
        // because the JBIListener implementation can access to the specific
        // component implementation which is done in the previous doInit()
        // call.
        this.processorManager = new JBIProcessorManager(this, this.runtimeConfigurationNotifier,
                this.logger);
        this.acceptorManager = new JBIAcceptorManager(this, this.processorManager,
                this.runtimeConfigurationNotifier, this.logger);
        this.asyncManager = new AsyncMessageManager(this.logger, this.processorManager);
        this.logger.info("Component initialized");
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.Component#isExchangeWithConsumerOkay
     * (javax.jbi.servicedesc.ServiceEndpoint,
     * javax.jbi.messaging.MessageExchange)
     */
    public boolean isExchangeWithConsumerOkay(final ServiceEndpoint endpoint,
            final MessageExchange exchange) {
        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.Component#isExchangeWithProviderOkay
     * (javax.jbi.servicedesc.ServiceEndpoint,
     * javax.jbi.messaging.MessageExchange)
     */
    public boolean isExchangeWithProviderOkay(final ServiceEndpoint endpoint,
            final MessageExchange exchange) {
        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.Component#resolveEndpointReference
     * (org.w3c.dom.DocumentFragment)
     */
    public ServiceEndpoint resolveEndpointReference(final DocumentFragment epr) {
        return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ComponentLifeCycle#shutDown()
     */
    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.Component#shutDown()
     */
    public final void shutDown() throws JBIException {
        this.logger.info("Shut down component...");
        try {
            this.logger.fine("Calling specific shut down...");
            this.doShutdown();
            this.logger.fine("Specific shut down done.");
        } finally {
            // remove unnecessary resources
            this.nativeWsdl = null;
            this.notificationService = null;
            if (this.componentServiceEndpoints != null) {
                this.componentServiceEndpoints.clear();
                this.componentServiceEndpoints = null;
            }
            this.channel = null;
            this.runtimeConfigurationNotifier = null;
            this.interceptors = null;
            this.interceptorManager = null;
            this.acceptorManager = null;
            this.processorManager = null;
            this.asyncManager = null;
        }
        this.logger.info("Component shut down");
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.Component#start()
     */
    public final void start() throws JBIException {
        this.logger.info("Start component...");
        if (this.channel == null) {
            this.channel = this.context.getDeliveryChannel();
        }
        this.registerRuntimeConfigMBean();
        this.registerInterceptorManagerMBean();
        this.processorManager.start();
        this.acceptorManager.start();
        this.asyncManager.start();
        this.activateServiceEndpointList();
        this.notificationService.start();
        this.logger.fine("Calling specific start...");
        this.doStart();
        this.logger.info("Component started");
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.Component#stop()
     */
    public final void stop() throws JBIException {
        this.logger.info("Stop component...");
        try {
            this.logger.fine("Calling specific stop...");
            this.doStop();
            this.logger.fine("Specific stop done.");
        } finally {
            this.notificationService.stop();
            this.deactivateServiceEndpointList();
            this.unregisterRuntimeConfigurationMBean();
            this.unregisterInterceptorManagerMBean();
            // stop first the acceptor Manager as it aggregates the worker
            // manager
            this.acceptorManager.stop();
            this.processorManager.stop();
            this.asyncManager.stop();
        }
        this.logger.info("Component stopped");
    }

    /**
     * Can be overridden to provide a specific service unit manager
     * 
     * @return the service unit manager used by this component.
     */
    protected abstract AbstractServiceUnitManager createServiceUnitManager();

    /**
     * A convenience method which can be overridden so that there's no need to
     * call super.doInit(). This method is called at the end of the component
     * classical initialization. When this method is called, the component's
     * delivery channel, context, logger and service unit manager are available.
     * 
     */
    protected void doInit() throws JBIException {
    }

    /**
     * A convenience method which can be overridden so that there's no need to
     * call super.doShutdown(). This method is called at the end of the
     * component classical shutdown.
     * 
     */
    protected void doShutdown() throws JBIException {
    }

    /**
     * A convenience method which can be overridden so that there's no need to
     * call super.doStart(). This method is called at the end of the component
     * classical start.
     * 
     */
    protected void doStart() throws JBIException {
    }

    /**
     * A convenience method which can be overridden so that there's no need to
     * call super.doStop(). This method is called at the end of the component
     * classical stop.
     * 
     */
    protected void doStop() throws JBIException {
    }

    /**
     * Can be overridden to return a custom resource bundle name
     * 
     * @return the resource bundle name or null if no resource bundle name has
     *         to be used (default value)
     */
    protected String getResourceBundleName() {
        return null;
    }

    /**
     * Store wsdl and then activate endpoint in the JBI context
     * 
     * @param serviceName
     * @param endpointName
     * @throws JBIException
     */
    public ServiceEndpoint activateEndpoint(final QName serviceName, final String endpointName)
            throws JBIException {

        // Fake EP for method getServicedescription...
        ServiceEndpointKey key = new ServiceEndpointKey(serviceName, endpointName);
        this.componentServiceEndpoints.put(key,
                new ComponentServiceEndpoint(serviceName, endpointName));
        final ServiceEndpoint se = this.context.activateEndpoint(serviceName, endpointName);
        this.componentServiceEndpoints.put(key, se);
        return se;
    }

    /**
     * Activate the internal service endpoint list managed directly by the
     * component. It contains the specific endpoints and the notification
     * producer endpoints
     * 
     * @throws JBIException
     */
    private void activateServiceEndpointList() throws JBIException {
        try {
            QName serviceName = null;
            String endpointName = null;
            List<Endpoint> endpointList = null;

            if (this.nativeWsdl != null) {
                endpointList = UtilFactory.getWSDLUtil().getEndpointList(
                        this.nativeWsdl.getDescription());
                if (endpointList != null) {
                    final Iterator<Endpoint> iterator = endpointList.iterator();
                    if (iterator != null) {
                        while (iterator.hasNext()) {
                            final Endpoint endpoint = iterator.next();
                            if (endpoint != null) {
                                serviceName = endpoint.getService().getQName();
                                endpointName = endpoint.getName();
                                this.activateEndpoint(serviceName, endpointName);
                            }
                        }
                    }
                }
            }
        } catch (final WSDLException e) {
            throw new JBIException("Unable to activate the endpoints managed by the component", e);
        }
    }

    /**
     * Deactivate the internal service endpoint list managed directly by the
     * component.
     * 
     * @throws JBIException
     */
    private void deactivateServiceEndpointList() throws JBIException {
        if (this.componentServiceEndpoints != null) {
            for (final ServiceEndpointKey key : this.componentServiceEndpoints.keySet()) {
                this.deactivateEndpoint(this.componentServiceEndpoints.get(key));
            }
            // finally clear remaining endpoints if all is ok...
            this.componentServiceEndpoints.clear();
        }
    }

    public void deactivateEndpoint(final ServiceEndpoint serviceEndpoint) throws JBIException {
        // remove from container
        this.context.deactivateEndpoint(serviceEndpoint);
        // remove from local cache
        this.componentServiceEndpoints.remove(new ServiceEndpointKey(serviceEndpoint
                .getServiceName(), serviceEndpoint.getEndpointName()));
    }

    /**
     * Load the interceptors if there is any.
     * 
     * @param configuration
     */
    private final Map<String, Interceptor> loadInterceptors() {
        final Map<String, Interceptor> interceptors = new HashMap<String, Interceptor>();

        if (this.jbiComponentConfiguration.getComponent().getComponentInterceptors() != null) {
            for (final ComponentInterceptor componentInterceptor : this.jbiComponentConfiguration
                    .getComponent().getComponentInterceptors().getInterceptor()) {
                try {
                    this.logger.info("Loading interceptor : " + componentInterceptor.getName());
                    Class<?> clazz = Thread.currentThread().getContextClassLoader().loadClass(
                            componentInterceptor.getClazz());
                    final Interceptor itc = (Interceptor) clazz.newInstance();

                    ConfigurationExtensions configurationExtensions;
                    if ((this.jbiComponentConfiguration.getComponent().getPropertiesFile() != null)
                            && !StringHelper.isNullOrEmpty(this.jbiComponentConfiguration
                                    .getComponent().getPropertiesFile().getValue())) {
                        configurationExtensions = new ConfigurationExtensions(componentInterceptor
                                .getParam(), this.jbiComponentConfiguration.getComponent()
                                .getPropertiesFile().getValue());
                    } else {
                        configurationExtensions = new ConfigurationExtensions(componentInterceptor
                                .getParam());
                    }

                    itc.init(this, componentInterceptor.getName(), componentInterceptor.isActive(),
                            configurationExtensions, this.logger);
                    interceptors.put(componentInterceptor.getName(), itc);
                } catch (final PEtALSCDKException e) {
                    this.logger.log(Level.WARNING, "Could not load or initialize interceptor '"
                            + componentInterceptor.getName() + "' from class '"
                            + componentInterceptor.getClazz()
                            + "', this interceptor is skipped. Cause : ", e);
                } catch (ClassNotFoundException e) {
                    this.logger.log(Level.WARNING, "Could not load or initialize interceptor '"
                            + componentInterceptor.getName() + "' from class '"
                            + componentInterceptor.getClazz()
                            + "', this interceptor is skipped. Cause : ", e);
                } catch (InstantiationException e) {
                    this.logger.log(Level.WARNING, "Could not load or initialize interceptor '"
                            + componentInterceptor.getName() + "' from class '"
                            + componentInterceptor.getClazz()
                            + "', this interceptor is skipped. Cause : ", e);
                } catch (IllegalAccessException e) {
                    this.logger.log(Level.WARNING, "Could not load or initialize interceptor '"
                            + componentInterceptor.getName() + "' from class '"
                            + componentInterceptor.getClazz()
                            + "', this interceptor is skipped. Cause : ", e);
                } catch (ClassCastException e) {
                    this.logger.log(Level.WARNING, "Could not load or initialize interceptor '"
                            + componentInterceptor.getName() + "' from class '"
                            + componentInterceptor.getClazz()
                            + "', this interceptor is skipped. Cause : ", e);
                }
            }
        }

        return interceptors;
    }

    /**
     * 
     * @throws JBIException
     */
    private void registerInterceptorManagerMBean() throws JBIException {
        final ObjectName mbeanName = this.context.getMBeanNames().createCustomComponentMBeanName(
                "interceptors_" + this.context.getComponentName());

        this.interceptorManager.setMBeanName(mbeanName);

        MBeanHelper
                .registerMBean(this.context.getMBeanServer(), mbeanName, this.interceptorManager);
    }

    /**
     * Register the {@link RuntimeConfigurationMBean}
     * 
     * @throws JBIException
     */
    private void registerRuntimeConfigMBean() throws JBIException {
        final ObjectName runtimeConfigurationMBeanName = this.context.getMBeanNames()
                .createCustomComponentMBeanName(
                        "runtime_configuration_" + this.context.getComponentName());

        this.runtimeConfigurationNotifier.setMBeanName(runtimeConfigurationMBeanName);

        MBeanHelper.registerMBean(this.context.getMBeanServer(), runtimeConfigurationMBeanName,
                this.runtimeConfigurationNotifier);
    }

    /**
     * 
     * @throws JBIException
     */
    private void unregisterInterceptorManagerMBean() throws JBIException {
        MBeanHelper.unregisterMBean(this.context.getMBeanServer(), this.interceptorManager
                .getMBeanName());
    }

    /**
     * Unregister the runtime configuration MBean
     * 
     * @throws JBIException
     */
    private void unregisterRuntimeConfigurationMBean() throws JBIException {
        MBeanHelper.unregisterMBean(this.context.getMBeanServer(),
                this.runtimeConfigurationNotifier.getMBeanName());
    }

    public org.ow2.petals.component.framework.api.notification.NotificationService getNotificationService() {
        return this.notificationService;
    }

    private static class ComponentServiceEndpoint implements ServiceEndpoint {
        private QName s = null;

        private String e = null;

        public ComponentServiceEndpoint(final QName serviceName, final String endpointName) {
            this.s = serviceName;
            this.e = endpointName;
        }

        public DocumentFragment getAsReference(final QName operationName) {
            return null;
        }

        public String getEndpointName() {
            return this.e;
        }

        public QName[] getInterfaces() {
            return null;
        }

        public QName getServiceName() {
            return this.s;
        }
    }
}
