/**
 * PETALS - PETALS Services Platform. Copyright (c) 2006 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework.listener;

import java.util.List;
import java.util.logging.Level;

import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.namespace.QName;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.xpath.XPathExpressionException;

import org.ow2.easywsdl.wsdl.api.WSDLException;
import org.ow2.petals.component.framework.api.Constants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.message.ExchangeImpl;
import org.ow2.petals.component.framework.process.async.AsyncContext;
import org.ow2.petals.component.framework.util.UtilFactory;
import org.w3c.dom.Document;

import com.ebmwebsourcing.wsstar.addressing.definition.WSAddressingFactory;
import com.ebmwebsourcing.wsstar.addressing.definition.api.EndpointReferenceType;
import com.ebmwebsourcing.wsstar.addressing.definition.api.WSAddressingException;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.WsnbConstants;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.NotificationMessageHolderType;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.Notify;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.Subscribe;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.SubscribeResponse;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.Unsubscribe;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.UnsubscribeResponse;
import com.ebmwebsourcing.wsstar.notification.definition.inout.WSNotificationReader;
import com.ebmwebsourcing.wsstar.notification.definition.inout.WSNotificationWriter;
import com.ebmwebsourcing.wsstar.notification.definition.utils.WSNotificationException;
import com.ebmwebsourcing.wsstar.notification.extension.utils.WSNotificationExtensionException;
import com.ebmwebsourcing.wsstar.notification.extension.utils.WsnSpecificTypeHelper;
import com.ebmwebsourcing.wsstar.notification.service.fault.WSNotificationFault;

/**
 * This class is used when a JBI message is received on the Component from the
 * JBI environment. <br>
 * Service Engine case : <br>
 * The goal of this class is to process a service with the JBI message
 * information.<br>
 * For instance, a ServiceUnit has been deployed on the component, defining a
 * <code>provides</code> section.<br>
 * This section defines an Endpoint <em>"myService/jbiMyServiceEndpoint"</em>
 * and some property extensions.<br>
 * A <em>"myService/jbiMyServiceEndpoint"</em> has been activated on the JBI
 * environment.<br>
 * <br>
 * When a JBI consumer sends in the JBI environment a message to the
 * <em>"myService/jbiMyServiceEndpoint"</em> endpoint, this component received
 * the message, and calls the <code>onJBIMessage(exchange)</code> method.<br>
 * Then, the method has to process the message according to optional property
 * extensions (extensions of the provides that has activated the endpoint).
 * 
 * Binding Component case : <br>
 * The goal of this class is to call an external service, related to this JBI
 * message. The address associated with the requested endpoint is handled by the
 * Extensions object of the provides that has activated the endpoint.<br>
 * For instance, a ServiceUnit has been deployed on the component, defining a
 * <code>provides</code> section.<br>
 * This section defines a relation between the external address
 * <em>"http://www.enterprise.com/webservices/myService"</em> (an address
 * attribute of the provides extensions), and an Endpoint
 * <em>"myService/jbiMyServiceEndpoint"</em>.<br>
 * A <em>"myService/jbiMyServiceEndpoint"</em> has been activated on the JBI
 * environment.<br>
 * <br>
 * When a JBI consumer sends in the JBI environment a message to the
 * <em>"myService/jbiMyServiceEndpoint"</em> endpoint, this component received
 * the message, and calls the <code>onJBIMessage(exchange)</code> method.<br>
 * Then, the method has to send the message to
 * <em>"http://www.enterprise.com/webservices/myService"</em>. <br/>
 * 
 * @author ofabre - EBM Websourcing
 * 
 */
public abstract class AbstractJBIListener extends AbstractListener {

    public AbstractJBIListener() {
        super();
    }

    /**
     * This method is invoked by the CDK when an asynchronous response is
     * received for the component. Its bound {@code AsyncContext} is set as
     * parameter.
     * <p />
     * This method MUST be overridden if asynchronous invocation are processed
     * by the component.
     * 
     * @param exchange
     * @param asyncContext
     */
    public boolean onAsyncJBIMessage(final Exchange asyncExchange, final AsyncContext asyncContext) {
        return false;
    }

    /**
     * This method is invoked by the CDK when an asynchronous Notification
     * response is received for the component. Its bound {@code AsyncContext} is
     * set as parameter.
     * <p />
     * Asynchronous invocation are processed by the component.
     * 
     * @param exchange
     * @param asyncContext
     */
    public boolean onAsyncNotificationMessage(final Exchange asyncExchange,
            final AsyncContext asyncContext) {
        // TODO : implement the asynchronous response process of a Notification
        // exchange
        return false;
    }

    /**
     * This method is invoked by the CDK when an asynchronous context has
     * expires. Its bound {@code AsyncContext} is set as parameter.
     * <p />
     * This method MUST be overridden if asynchronous invocation are processed
     * by the component.
     * 
     * @param exchange
     * @param asyncContext
     */
    public boolean onExpiredAsyncJBIMessage(final Exchange originalExchange,
            final AsyncContext asyncContext) {
        return false;
    }

    /**
     * This method is invoked by the CDK when a Notification asynchronous
     * context has expires. Its bound {@code AsyncContext} is set as parameter.
     * <p />
     * Asynchronous invocation are processed by the component.
     * 
     * @param exchange
     * @param asyncContext
     */
    public boolean onExpiredAsyncNotificationMessage(final Exchange originalExchange,
            final AsyncContext asyncContext) {
        // TODO : implement the asynchronous request process of a Notification
        // exchange expired
        return false;
    }

    /**
     * <p>
     * This method is used to process a JBI message received on the Component
     * from the JBI environment.
     * </p>
     * 
     * <p>
     * This method can't throws Exceptions. You must managed Exception yourself
     * by using {@link Exchange} methods :<br/>
     * - setFault(Exception) : create a {@link Fault} with the given Exception
     * and set it in the {@link MessageExchange}<br/>
     * - setError(Exception) : set Error part of the {@link MessageExchange} and
     * set {@link MessageExchange} status to error. Exceptions handled by
     * {@link MessageExchange}s must be in the container classpath.
     * </p>
     * 
     * <p>
     * If the methods return {@code true}, according to the pattern use cases,
     * message can be terminated and sent back by the CDK :<br/>
     * - InOnly : DONE status is automatically set at the end of the message
     * exchange process.<br/>
     * - RobustInOnly : DONE status is automatically set if no Fault found in
     * exchange at the end of the message exchange process.<br/>
     * - InOut : DONE status is never set (forbidden).<br/>
     * - InOptOut : DONE status is automatically set if no Fault or Out message
     * found in exchange at the end of the message exchange process.
     * 
     * @param exchange
     *            the message exchange to process, wrapped to add some utility
     *            methods. Not null
     * 
     * @return {@code true} if the message must be terminated and sent back
     *         automatically by the CDK. The interceptors and notifications are
     *         activated.<br /> {@code false} if you you want yourself terminate the
     *         process and not use the
     *         {@link org.ow2.petals.component.framework.process.MessageExchangeProcessor}
     * 
     * @throws RuntimeException
     *             RuntimeException will be handle by the MessageExchange as
     *             Error part. So handled {@link RuntimeException}s must be in
     *             the classpath of your Petals node(s)
     */
    public abstract boolean onJBIMessage(Exchange exchange);

    /**
     * <p>
     * This method is used to process a Notification message received on the
     * Component from the JBI environment. Today, a Notification message is
     * provided by a JBI exchange
     * </p>
     * <p>
     * If the methods return {@code true}, according to the pattern use cases,
     * message can be terminated and sent back by the CDK :<br/>
     * - InOnly : DONE status is automatically set at the end of the message
     * exchange process.<br/>
     * - RobustInOnly : DONE status is automatically set if no Fault found in
     * exchange at the end of the message exchange process.<br/>
     * - InOut : DONE status is never set (forbidden).<br/>
     * - InOptOut : DONE status is automatically set if no Fault or Out message
     * found in exchange at the end of the message exchange process.
     * 
     * @param exchange
     *            the message exchange to process, wrapped to add some utility
     *            methods. Not null
     * 
     * @return {@code true} if the message must be terminated and sent back
     *         automatically by the CDK. {@code false} if you are using
     *         asynchronous invocations and you want to terminate the process
     * 
     * @throws RuntimeException
     *             RuntimeException will be handle by the MessageExchange as
     *             Error part. So handled {@link RuntimeException}s must be in
     *             the classpath of your Petals node(s)
     */
    public boolean onNotificationMessage(final Exchange exchange) {
        boolean response = true;
        NormalizedMessage normalizedMessage = null;
        Document document = null;
        EndpointReferenceType endpointReferenceType = null;
        String address = null;

        try {
            if (this.getLogger().isLoggable(Level.FINE)) {
                for (final QName intf : exchange.getEndpoint().getInterfaces()) {
                    this.getLogger().fine("notification endpoint interface: " + intf);
                }
                this.getLogger()
                .fine(
                        "notification endpoint service: "
                        + exchange.getEndpoint().getServiceName());
                this.getLogger().fine(
                        "notification endpoint name: " + exchange.getEndpoint().getEndpointName());
                this.getLogger()
                .fine("notification operation name: " + exchange.getOperationName());
            }
            if (exchange.isActiveStatus()) {
                if (exchange.getFault() != null) {
                    if (UtilFactory.getExchangeUtil().isPetalsException(exchange.getFault())) {
                        this.getLogger().warning(
                                "notification technical fault message content: "
                                + UtilFactory.getSourceUtil().createString(
                                        exchange.getFault().getContent()));
                    } else {
                        this.getLogger().warning(
                                "notification business fault message content: "
                                + UtilFactory.getSourceUtil().createString(
                                        exchange.getFault().getContent()));
                    }
                } else if (this.component.getNotificationBrokerController() == null) {
                    this
                    .getLogger()
                    .warning(
                            "Unable to analyze the notification : the notification broker controller isn't initialized");
                } else {
                    // TODO : found a solution to call the onJBIExchange from
                    // another class and externalize this source
                    normalizedMessage = exchange.getInMessage();
                    if (WsnbConstants.SUBSCRIBE_QNAME.equals(exchange.getOperation())) {
                        document = UtilFactory.getSourceUtil().createDocument(
                                normalizedMessage.getContent());
                        Subscribe subscribe = WSNotificationReader.getInstance().readSubscribe(
                                document);
                        address = subscribe.getConsumerReference().getAddress();
                        if (address == null
                                || this.component.getNotificationBrokerController()
                                .isInternalNotification(address)) {
                            // call the producer
                            final SubscribeResponse subscribeResponse = this.component
                            .getNotificationBrokerController().getNotificationProducer()
                            .subscribe(subscribe);
                            // set the response
                            document = WSNotificationWriter.getInstance().writeSubscribeResponse(
                                    subscribeResponse);
                            normalizedMessage = exchange.getOutMessage();
                            normalizedMessage.setContent(UtilFactory.getSourceUtil()
                                    .createStreamSource(document));
                            exchange.setOutMessage(normalizedMessage);
                        } else {
                            ((ExchangeImpl) exchange).setCheckRole(false);
                            // WS-Addressing
                            endpointReferenceType = WSAddressingFactory.getInstance()
                            .newEndpointReferenceType();
                            if (normalizedMessage.getProperty(Constants.WSStar.ADDRESSING_KEY) == null) {
                                endpointReferenceType.setAddress(address);
                                document = WSAddressingFactory.getInstance()
                                .newWSAddressingWriter().getDocument(endpointReferenceType);
                                normalizedMessage.setProperty(Constants.WSStar.ADDRESSING_KEY,
                                        document);
                            }
                            // transform the Subscribe
                            subscribe = this.component.getNotificationBrokerController()
                            .getTransformer().transformRequestOnProducer(subscribe);
                            document = WSNotificationWriter.getInstance().writeSubscribe(subscribe);
                            normalizedMessage.setContent(UtilFactory.getSourceUtil()
                                    .createStreamSource(document));
                            exchange.setInMessage(normalizedMessage);
                            response = this.onJBIMessage(exchange);
                            if (exchange.isActiveStatus() && exchange.getOutMessage() != null) {
                                // transform the Subscribe Response
                                normalizedMessage = exchange.getOutMessage();
                                document = UtilFactory.getSourceUtil().createDocument(
                                        normalizedMessage.getContent());
                                SubscribeResponse subscribeResponse = WSNotificationReader
                                .getInstance().readSubscribeResponse(document);
                                subscribeResponse = this.component
                                .getNotificationBrokerController().getTransformer()
                                .transformResponseOnProducer(subscribeResponse);
                                document = WSNotificationWriter.getInstance()
                                .writeSubscribeResponse(subscribeResponse);
                                normalizedMessage.setContent(UtilFactory.getSourceUtil()
                                        .createStreamSource(document));
                                exchange.setOutMessage(normalizedMessage);
                            }
                            ((ExchangeImpl) exchange).setCheckRole(true);
                        }
                    } else if (WsnbConstants.NOTIFY_QNAME.equals(exchange.getOperation())) {
                        document = UtilFactory.getSourceUtil().createDocument(
                                normalizedMessage.getContent());
                        Notify notify = WSNotificationReader.getInstance().readNotify(document);
                        final List<NotificationMessageHolderType> notificationMessageList = notify
                        .getNotificationMessage();
                        if (notificationMessageList == null || notificationMessageList.size() != 1) {
                            exchange.setError(new Exception(
                                    "The CDK need one and only one notification message"));
                        } else if (notificationMessageList.get(0).getSubscriptionReference() != null) {
                            address = notificationMessageList.get(0).getSubscriptionReference()
                            .getAddress();
                        }
                        if (address == null
                                || this.component.getNotificationBrokerController()
                                .isInternalNotification(address)) {
                            // call the consumer
                            this.component.getNotificationBrokerController()
                            .getNotificationConsumer().notify(notify);
                        } else {
                            ((ExchangeImpl) exchange).setCheckRole(false);
                            // WS-Addressing
                            endpointReferenceType = WSAddressingFactory.getInstance()
                            .newEndpointReferenceType();
                            if (normalizedMessage.getProperty(Constants.WSStar.ADDRESSING_KEY) == null) {
                                endpointReferenceType.setAddress(address);
                                document = WSAddressingFactory.getInstance()
                                .newWSAddressingWriter().getDocument(endpointReferenceType);
                                normalizedMessage.setProperty(Constants.WSStar.ADDRESSING_KEY,
                                        document);
                            }
                            // transform the Notify
                            notify = this.component.getNotificationBrokerController()
                            .getTransformer().transformRequestOnProducer(notify);
                            document = WSNotificationWriter.getInstance().writeNotify(notify);
                            normalizedMessage.setContent(UtilFactory.getSourceUtil()
                                    .createStreamSource(document));
                            exchange.setInMessage(normalizedMessage);
                            response = this.onJBIMessage(exchange);
                            ((ExchangeImpl) exchange).setCheckRole(true);
                        }
                    } else if (WsnbConstants.UNSUBSCRIBE_QNAME.equals(exchange.getOperation())) {
                        document = UtilFactory.getSourceUtil().createDocument(
                                normalizedMessage.getContent());
                        Unsubscribe unsubscribe = WSNotificationReader.getInstance()
                        .readUnsubscribe(document);
                        final EndpointReferenceType endpointReference = WsnSpecificTypeHelper
                        .getEndpointReference(unsubscribe);
                        if (endpointReference != null) {
                            address = endpointReference.getAddress();
                        }
                        if (address == null
                                || this.component.getNotificationBrokerController()
                                .isInternalNotification(address)) {
                            // call the producer
                            final UnsubscribeResponse unsubscribeResponse = this.component
                            .getNotificationBrokerController().getNotificationProducer()
                            .unsubscribe(unsubscribe);
                            // set the response
                            document = WSNotificationWriter.getInstance().writeUnsubscribeResponse(
                                    unsubscribeResponse);
                            normalizedMessage = exchange.getOutMessage();
                            normalizedMessage.setContent(UtilFactory.getSourceUtil()
                                    .createStreamSource(document));
                            exchange.setOutMessage(normalizedMessage);
                        } else {
                            ((ExchangeImpl) exchange).setCheckRole(false);
                            // WS-Addressing
                            endpointReferenceType = WSAddressingFactory.getInstance()
                            .newEndpointReferenceType();
                            if (normalizedMessage.getProperty(Constants.WSStar.ADDRESSING_KEY) == null) {
                                endpointReferenceType.setAddress(address);
                                document = WSAddressingFactory.getInstance()
                                .newWSAddressingWriter().getDocument(endpointReferenceType);
                                normalizedMessage.setProperty(Constants.WSStar.ADDRESSING_KEY,
                                        document);
                            }
                            // transform the Unsubscribe
                            unsubscribe = this.component.getNotificationBrokerController()
                            .getTransformer().transformRequestOnProducer(unsubscribe);
                            document = WSNotificationWriter.getInstance().writeUnsubscribe(
                                    unsubscribe);
                            normalizedMessage.setContent(UtilFactory.getSourceUtil()
                                    .createStreamSource(document));
                            exchange.setInMessage(normalizedMessage);
                            response = this.onJBIMessage(exchange);
                            // transform the Unsubscribe Response: nothing to do
                            // now
                            ((ExchangeImpl) exchange).setCheckRole(true);
                        }
                    } else if (WsnbConstants.GET_CURRENT_MESSAGE_QNAME.equals(exchange
                            .getOperation())) {
                        // TODO: implements the management
                    } else if (WsnbConstants.RENEW_QNAME.equals(exchange.getOperation())) {
                        // TODO: implements the management
                    } else {
                        exchange
                        .setError(new Exception(
                                "unable to identify an operation of the WS-Notification specifications"));
                    }
                }
            }
        } catch (final PEtALSCDKException e) {
            exchange.setError(new Exception(e));
        } catch (final WSNotificationException e) {
            exchange.setError(new Exception(e));
        } catch (final WSAddressingException e) {
            exchange.setError(new Exception(e));
        } catch (final MessagingException e) {
            exchange.setError(new Exception(e));
        } catch (final WSDLException e) {
            exchange.setError(new Exception(e));
        } catch (final WSNotificationExtensionException e) {
            exchange.setError(new Exception(e));
        } catch (final XPathExpressionException e) {
            exchange.setError(new Exception(e));
        } catch (final TransformerFactoryConfigurationError e) {
            exchange.setError(new Exception(e));
        } catch (final TransformerException e) {
            exchange.setError(new Exception(e));
        } catch (final ParserConfigurationException e) {
            exchange.setError(new Exception(e));
        } catch (final WSNotificationFault f) {
            try {
                exchange.setFault(f);
            } catch (final MessagingException e) {
                exchange.setError(new Exception(e));
            }
        }
        return response;
    }
}
