/**
 * PETALS: PETALS Services Platform Copyright (C) 2009 EBM WebSourcing
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 * 
 * Initial developer(s): EBM WebSourcing
 */

package org.ow2.petals.component.framework.notification;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.messaging.MessagingException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.ow2.easywsdl.wsdl.api.Endpoint;
import org.ow2.easywsdl.wsdl.api.WSDLException;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.api.Constants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.notification.AsyncSender;
import org.ow2.petals.component.framework.api.notification.NotificationBrokerController;
import org.ow2.petals.component.framework.api.notification.NotificationController;
import org.ow2.petals.component.framework.api.notification.Sender;
import org.ow2.petals.component.framework.api.notification.StatusListener;
import org.ow2.petals.component.framework.jbidescriptor.generated.Component;
import org.ow2.petals.component.framework.util.UtilFactory;

import com.ebmwebsourcing.wsstar.addressing.definition.api.WSAddressingException;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.SubscribeResponse;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.UnsubscribeResponse;
import com.ebmwebsourcing.wsstar.notification.definition.brokerednotification.api.DestroyRegistrationResponse;
import com.ebmwebsourcing.wsstar.notification.definition.brokerednotification.api.RegisterPublisherResponse;
import com.ebmwebsourcing.wsstar.notification.definition.utils.WSNotificationException;
import com.ebmwebsourcing.wsstar.notification.extension.utils.WSNotificationExtensionException;

/**
 * The notification service is in charge on handling notification related tasks
 * for the current component. Note that the service can be activated at runtime,
 * so the lifecyle methods must be able to handle this runtime activation...
 * 
 * @author chamerling - eBM WebSourcing
 * 
 */
public class NotificationService implements
        org.ow2.petals.component.framework.api.notification.NotificationService {

    private final AbstractComponent component;

    private final NotificationContext notificationContext;

    private NotificationStorageService notificationStorageService;

    private Set<StatusListener> statusListeners;

    /**
     * The broker controller of notification of the system
     * 
     * @deprecated
     */
    @Deprecated
    protected NotificationBrokerController notificationBrokerController;

    /**
     * Cache of notification endpoints
     */
    private Map<String, ServiceEndpoint> endpointCache;

    /**
     * The controller of notification of the system. FIXME : Not used for now,
     * once used, need to report all the work done for the broker controller.
     */
    private NotificationController notificationController;

    /**
     * The registration uuid for the cdk notifications
     */
    private String registrationUuid = null;

    private final Logger logger;

    private boolean active;

    /**
     * The sender used to send notifications through the bus
     */
    private WrappedSender sender;

    /**
     * @throws MessagingException
     * @throws WSNotificationException
     * 
     */
    public NotificationService(final AbstractComponent component) throws WSNotificationException,
            MessagingException {
        this.component = component;
        this.notificationContext = new NotificationContext();
        this.logger = this.component.getLogger();
        this.statusListeners = new HashSet<StatusListener>();
    }

    /**
     * {@inheritDoc}
     */
    public void init() throws PEtALSCDKException {
        this.endpointCache = new HashMap<String, ServiceEndpoint>();
        Component componentDescriptor = this.component.getComponentConfiguration();
        if (componentDescriptor.getNotifRetryPolicyMax() != null) {
            this.notificationContext.setMaxDelay(componentDescriptor.getNotifRetryPolicyMax()
                    .getValue());
        }
        if (componentDescriptor.getNotifRetryPolicyMin() != null) {
            this.notificationContext.setMinDelay(componentDescriptor.getNotifRetryPolicyMin()
                    .getValue());
        }
        if (componentDescriptor.getNotifRetryPolicyFactor() != null) {
            this.notificationContext.setDelayFactor(componentDescriptor.getNotifRetryPolicyFactor()
                    .getValue());
        }
        if (componentDescriptor.getNotifRetryPolicyFactor() != null) {
            this.notificationContext.setNbRetry(componentDescriptor.getNotifRetryPolicyNb()
                    .getValue());
        }

        this.notificationStorageService = new NotificationStorageService(this.component
                .getContext().getWorkspaceRoot(), this.sender, this.notificationContext,
                this.logger);
        this.notificationStorageService.init();
        this.addListener(this.notificationStorageService);

        try {
            this.sender = new WrappedSender(this.component.getContext(), this,
                    this.notificationStorageService, this.notificationContext, this.logger);
            this.sender.init();
        } catch (final WSNotificationException e1) {
            // TODO Auto-generated catch block
            e1.printStackTrace();
        } catch (final MessagingException e1) {
            // TODO Auto-generated catch block
            e1.printStackTrace();
        }

        this.active = (this.component != null)
                && (this.component.getComponentConfiguration().getNotifications() != null)
                && this.component.getComponentConfiguration().getNotifications().isValue();

        try {
            this.notificationBrokerController = new ComponentNotificationBrokerController(
                    this.logger, this.component.getContext(), this);
        } catch (final WSNotificationException e) {
            throw new PEtALSCDKException(e);
        } catch (final WSDLException e) {
            throw new PEtALSCDKException(e);
        } catch (final WSAddressingException e) {
            throw new PEtALSCDKException(e);
        } catch (final WSNotificationExtensionException e) {
            throw new PEtALSCDKException(e);
        } catch (final MessagingException e) {
            throw new PEtALSCDKException(e);
        }
    }

    /**
     * The notification service is started on component start lifecycle step. It
     * can also be launched on management operation at runtime.
     * 
     * {@inheritDoc}
     */
    public void start() throws PEtALSCDKException {
        this.logger.fine("Starting the notification service");
        this.notificationStorageService.start();
        this.sender.start();
        if (!this.active) {
            this.logger.fine("Notification feature is not active");
            return;
        }
        this.activate();
    }

    /**
     * @throws PEtALSCDKException
     * 
     */
    public void stop() throws PEtALSCDKException {
        this.logger.fine("Stopping the notification service");
        this.notificationStorageService.stop();
        this.unactivate();
        this.unregisterAsPublisher();
        this.sender.stop();
    }

    public NotificationBrokerController getNotificationBrokerController() {
        return this.notificationBrokerController;
    }

    public NotificationController getNotificationController() {
        return this.notificationController;
    }

    /**
     * 
     */
    protected void unregisterAsPublisher() {
        if (!this.active) {
            this.logger.fine("Notifications are not active, do not unregister on broker");
            return;
        }

        if (this.notificationController != null) {
            try {
                this.notificationController.destroyRegistration(this.registrationUuid);
            } catch (final Exception e) {
                this.logger
                        .warning("unable to destroy the registration of the publisher: the broker will try to send me notifications");
                this.logger.fine("detail on the error: " + e.getMessage());
                this.notificationController = null;
            }
        }
    }

    /**
     * @throws PEtALSCDKException
     */
    protected void unregisterFromBroker() throws PEtALSCDKException {
        if (!this.active) {
            this.logger.fine("Notifications are not active, do not unregister on broker");
            return;
        }

        if (this.notificationBrokerController != null) {
            try {
                this.notificationBrokerController.destroyRegistration();
            } catch (final WSNotificationException e) {
                this.logger
                        .warning("unable to destroy the registration of the publisher: the broker will try to send me notifications");
                this.logger.fine("detail on the error: " + e.getMessage());
                this.notificationBrokerController = null;
            } catch (final WSAddressingException e) {
                this.logger
                        .warning("unable to destroy the registration of the publisher: the broker will try to send me notifications");
                this.logger.fine("detail on the error: " + e.getMessage());
                this.notificationBrokerController = null;
            } catch (final WSNotificationExtensionException e) {
                this.logger
                        .warning("unable to destroy the registration of the publisher: the broker will try to send me notifications");
                this.logger.fine("detail on the error: " + e.getMessage());
                this.notificationBrokerController = null;
            } catch (final MessagingException e) {
                this.logger
                        .warning("unable to destroy the registration of the publisher: the broker will try to send me notifications");
                this.logger.fine("detail on the error: " + e.getMessage());
                this.notificationBrokerController = null;
            }
        }
    }

    /**
     * 
     */
    protected void registerNotificationPublisher() {
        if (this.notificationController != null) {
            try {
                this.registrationUuid = this.notificationController.registerPublisher(
                        Constants.WSStar.Topic.NAMESPACE_URI,
                        "component/cdk//*[@wstop:topic='true']");
            } catch (final Exception e) {
                this.logger
                        .warning("unable to register the publisher: no notification will be sent");
                this.logger.fine("detail on the error: " + e.getMessage());
                this.notificationController = null;
            }
        }
    }

    /**
     * @throws PEtALSCDKException
     */
    protected void registerNotificationPublisherOnBroker() throws PEtALSCDKException {
        if (this.notificationBrokerController != null) {
            try {
                // Do it in a spearate thread...

                this.notificationBrokerController.registerPublisher();
            } catch (final WSNotificationException e) {
                this.logger
                        .warning("unable to register the publisher: no notification will be sent");
                this.logger.fine("detail on the error: " + e.getMessage());
                this.notificationBrokerController = null;
            } catch (final WSAddressingException e) {
                this.logger
                        .warning("unable to register the publisher: no notification will be sent");
                this.logger.fine("detail on the error: " + e.getMessage());
                this.notificationBrokerController = null;
            } catch (final WSNotificationExtensionException e) {
                this.logger
                        .warning("unable to register the publisher: no notification will be sent");
                this.logger.fine("detail on the error: " + e.getMessage());
                this.notificationBrokerController = null;
            } catch (final MessagingException e) {
                this.logger
                        .warning("unable to register the publisher: no notification will be sent");
                this.logger.fine("detail on the error: " + e.getMessage());
                this.notificationBrokerController = null;
            }
        }
    }

    /**
     * @return the notificationContext
     */
    public NotificationContext getNotificationContext() {
        return this.notificationContext;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Sender getSender() {
        return this.sender;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public AsyncSender getAsyncSender(StatusListener listener) {
        return new org.ow2.petals.component.framework.notification.AsyncSender(getSender(),
                listener);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public synchronized void setActive(final boolean status) throws PEtALSCDKException {
        if (status && !active) {
            this.activate();
        } else if (active && !status) {
            this.unactivate();
        } else {
            logger.fine("Incoherent setActive change");
        }
        if (this.active != status) {
            this.active = status;
        }
    }

    /**
     * Really activate the notification feature
     * 
     * @throws PEtALSCDKException
     */
    private void activate() throws PEtALSCDKException {
        this.activateEndpoints();
        this.registerNotificationPublisherOnBroker();
        // FIXME : activate when it will be integrated...
        // this.registerNotificationPublisher();
    }

    private void activateEndpoints() throws PEtALSCDKException {
        // TODO remove the controller without managers
        // FIXME : This MUST be done when we are really ready to process
        // notifications...
        QName serviceName = null;
        String endpointName = null;
        ServiceEndpoint se = null;
        List<Endpoint> endpointList = null;

        try {

            if (this.getNotificationBrokerController() != null) {
                if (this.getNotificationBrokerController().getNotificationProducer() != null) {
                    endpointList = UtilFactory.getWSDLUtil().getEndpointList(
                            this.getNotificationBrokerController().getNotificationProducer()
                                    .getWsdl().getDescription());
                    if (endpointList != null) {
                        final Iterator<Endpoint> iterator = endpointList.iterator();
                        if (iterator != null) {
                            while (iterator.hasNext()) {
                                final Endpoint endpoint = iterator.next();
                                if (endpoint != null) {
                                    serviceName = endpoint.getService().getQName();
                                    endpointName = endpoint.getName();
                                    se = this.component.activateEndpoint(serviceName, endpointName);
                                    endpointCache.put(endpointName, se);
                                }
                            }
                        }
                    }
                }
                if (this.getNotificationBrokerController().getNotificationConsumer() != null) {
                    endpointList = UtilFactory.getWSDLUtil().getEndpointList(
                            this.getNotificationBrokerController().getNotificationConsumer()
                                    .getWsdl().getDescription());
                    if (endpointList != null) {
                        final Iterator<Endpoint> iterator = endpointList.iterator();
                        if (iterator != null) {
                            while (iterator.hasNext()) {
                                final Endpoint endpoint = iterator.next();
                                if (endpoint != null) {
                                    serviceName = endpoint.getService().getQName();
                                    endpointName = endpoint.getName();
                                    se = this.component.activateEndpoint(serviceName, endpointName);
                                    endpointCache.put(endpointName, se);
                                }
                            }
                        }
                    }
                }
            }
            if (this.getNotificationController() != null) {
                if (this.getNotificationController().getProducerWsdl() != null) {
                    endpointList = UtilFactory.getWSDLUtil().getEndpointList(
                            this.getNotificationController().getProducerWsdl().getDescription());
                    if (endpointList != null) {
                        final Iterator<Endpoint> iterator = endpointList.iterator();
                        if (iterator != null) {
                            while (iterator.hasNext()) {
                                final Endpoint endpoint = iterator.next();
                                if (endpoint != null) {
                                    serviceName = endpoint.getService().getQName();
                                    endpointName = endpoint.getName();
                                    se = this.component.activateEndpoint(serviceName, endpointName);
                                    endpointCache.put(endpointName, se);
                                }
                            }
                        }
                    }
                }
                if (this.getNotificationController().getConsumerWsdl() != null) {
                    endpointList = UtilFactory.getWSDLUtil().getEndpointList(
                            this.getNotificationController().getConsumerWsdl().getDescription());
                    if (endpointList != null) {
                        final Iterator<Endpoint> iterator = endpointList.iterator();
                        if (iterator != null) {
                            while (iterator.hasNext()) {
                                final Endpoint endpoint = iterator.next();
                                if (endpoint != null) {
                                    serviceName = endpoint.getService().getQName();
                                    endpointName = endpoint.getName();
                                    se = this.component.activateEndpoint(serviceName, endpointName);
                                    endpointCache.put(endpointName, se);
                                }
                            }
                        }
                    }
                }
            }
        } catch (final Exception e) {
            throw new PEtALSCDKException(
                    "Unable to activate the notification endpoints managed by the component", e);
        }
    }

    /**
     * 
     */
    private void unactivate() throws PEtALSCDKException {
        this.unregisterFromBroker();
        // this.unregisterAsPublisher();
        this.deactivateEndpoints();
    }

    /**
     * Deactivate the notification endpoints
     * 
     * @throws PEtALSCDKException
     */
    private void deactivateEndpoints() throws PEtALSCDKException {
        for (ServiceEndpoint serviceEndpoint : endpointCache.values()) {
            this.deactivateEndpoint(serviceEndpoint);
        }
    }

    /**
     * Deactivate a notification endpoint
     * 
     * @param serviceEndpoint
     * @throws PEtALSCDKException
     */
    private void deactivateEndpoint(ServiceEndpoint serviceEndpoint) throws PEtALSCDKException {
        try {
            this.component.deactivateEndpoint(serviceEndpoint);
        } catch (JBIException e) {
            throw new PEtALSCDKException("Can not deactivate notification endpoint", e);
        }
    }

    /**
     * @return the active
     */
    public boolean isActive() {
        return this.active;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void addListener(StatusListener listener) {
        if (listener != null) {
            this.statusListeners.add(listener);
        }
    }

    /**
     * The service which registers succefully the component MUST call this
     * method once done. {@inheritDoc}
     */
    @Override
    public void onRegistered(RegisterPublisherResponse response) {
        // no special order for now...
        for (StatusListener listener : this.statusListeners) {
            try {
                listener.onRegistered(response);
            } catch (PEtALSCDKException e) {
                this.logger.log(Level.WARNING, "Listener raised an exception on onRegistered call",
                        e);
            }
        }
    }

    /**
     * The service which unregisters succefully the component MUST call this
     * method once done. {@inheritDoc}
     */
    @Override
    public void onUnregistered(DestroyRegistrationResponse response) {
        // no special order for now...
        for (StatusListener listener : this.statusListeners) {
            try {
                listener.onUnregistered(response);
            } catch (PEtALSCDKException e) {
                this.logger.log(Level.WARNING,
                        "Listener raised an exception on onUnregistered call", e);
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void onSubscribed(SubscribeResponse response) throws PEtALSCDKException {
        // no special order for now...
        for (StatusListener listener : this.statusListeners) {
            try {
                listener.onSubscribed(response);
            } catch (PEtALSCDKException e) {
                this.logger.log(Level.WARNING, "Listener raised an exception on onSubscribed call",
                        e);
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void onUnsubscribed(UnsubscribeResponse response)
            throws PEtALSCDKException {
        // no special order for now...
        for (StatusListener listener : this.statusListeners) {
            try {
                listener.onUnsubscribed(response);
            } catch (PEtALSCDKException e) {
                this.logger.log(Level.WARNING,
                        "Listener raised an exception on onUnsubscribed call", e);
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void onException(Exception e) {
        // no special order for now...
        for (StatusListener listener : this.statusListeners) {
            listener.onException(e);
        }
    }
}
