/**
 * PETALS: PETALS Services Platform Copyright (C) 2009 EBM WebSourcing
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 * 
 * Initial developer(s): EBM WebSourcing
 */

package org.ow2.petals.component.framework.notification;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.notification.LifeCyle;
import org.ow2.petals.component.framework.api.notification.StatusListener;
import org.ow2.petals.component.framework.util.XMLUtil;
import org.w3c.dom.Document;

import com.ebmwebsourcing.wsstar.addressing.definition.api.EndpointReferenceType;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.Notify;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.SubscribeResponse;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.UnsubscribeResponse;
import com.ebmwebsourcing.wsstar.notification.definition.brokerednotification.api.DestroyRegistrationResponse;
import com.ebmwebsourcing.wsstar.notification.definition.brokerednotification.api.RegisterPublisherResponse;
import com.ebmwebsourcing.wsstar.notification.definition.inout.WSNotificationReader;
import com.ebmwebsourcing.wsstar.notification.definition.inout.WSNotificationWriter;
import com.ebmwebsourcing.wsstar.notification.definition.utils.WSNotificationException;

/**
 * Service used to store notifications which can not be sent. The notifications
 * will be sent ASAP by a processor...
 * 
 * @author chamerling - eBM WebSourcing
 * 
 */
public class NotificationStorageService implements LifeCyle, StatusListener {

    static final String PERSIST_FOLDER = "notifications";

    static final String TRASH = "trash";

    static final String POISON = "poison";

    protected static final long RESEND_DELAY = 10;

    private STATE state = STATE.UNKNOWN;

    private final String rootFolder;

    /**
     * Base folder to persist files
     */
    private File folder;

    /**
     * A trash folder used to move notification which can not be deleted...
     */
    private File trashFolder;

    private final Logger logger;

    private final AtomicLong fileCounter;

    private final org.ow2.petals.component.framework.api.notification.Sender sender;

    private ExecutorService worker;

    private final BlockingQueue<String> blockingQueue;

    private volatile boolean tryToSendNotification;

    private NotificationContext notificationContext;

    /**
     * 
     */
    public NotificationStorageService(final String folder,
            final org.ow2.petals.component.framework.api.notification.Sender sender,
            NotificationContext notificationContext, final Logger logger) {
        this.rootFolder = folder;
        this.logger = logger;
        this.sender = sender;
        this.notificationContext = notificationContext;
        this.fileCounter = new AtomicLong();
        this.blockingQueue = new LinkedBlockingQueue<String>();
    }

    /**
     * Store a notification
     * 
     * @throws PEtALSCDKException
     */
    void store(com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.Notify notify)
            throws PEtALSCDKException {
        if (this.state != STATE.INIT) {
            throw new PEtALSCDKException("Invalid state to store notification " + this.state);
        }

        this.logger.info("Storing notification");
        try {
            final Document document = WSNotificationWriter.getInstance().writeNotify(notify);
            final String id = this.getFileName();
            final File persistFile = new File(this.folder, id);
            if (persistFile.exists()) {
                persistFile.createNewFile();
            }
            XMLUtil.writeDocument(document, new FileOutputStream(persistFile));
            this.blockingQueue.add(id);
        } catch (final WSNotificationException e) {
            e.printStackTrace();
        } catch (final Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * @param notify
     * @return
     */
    private String getFileName() {
        // get a counter and the date (can be useful some day...)
        return this.fileCounter.incrementAndGet() + "-" + System.currentTimeMillis();
    }

    /**
     * 
     */
    public void init() {
        if (this.folder == null) {
            this.logger.fine("Creating notification persistance folder");
            this.folder = new File(this.rootFolder, PERSIST_FOLDER);
            this.folder.mkdirs();
        }

        if (this.trashFolder == null) {
            this.logger.fine("Creating notification persistance backup folder");
            this.trashFolder = new File(this.folder, TRASH);
            this.trashFolder.mkdirs();
        }

        // TODO : load all the persisted notification from the folder in the
        // process queue so that the processor will send all the remaining files
        // on restart
        this.worker = Executors.newSingleThreadExecutor();
        this.state = STATE.INIT;
    }

    /**
     * @param id
     * @return
     */
    protected long getFileCreationDate(final String id) {
        return Long.parseLong(id.substring(id.indexOf('-') + 1));
    }

    /**
     * @param f
     */
    protected void moveToTrash(final File f) {
        // NOTE : should do it with Apache commons IO when available in the CDK
        if (f.exists() && f.isFile()) {
            f.renameTo(new File(this.trashFolder, f.getName()));
        }
    }

    /**
     * 
     */
    public void start() {

        this.tryToSendNotification = false;
        this.state = STATE.STARTED;
    }

    /**
     * Process the notifications which have been persisted. This is generally
     * called when the component is finally registered as publisher on the
     * broker...
     */
    private void processNotifications() throws PEtALSCDKException {
        if (this.state != STATE.STARTED) {
            throw new PEtALSCDKException("Can not process notifications in state " + this.state);
        }
        this.tryToSendNotification = true;

        this.worker.execute(new Runnable() {
            @Override
            public void run() {
                NotificationStorageService.this.logger
                        .info("Storage Service is now running to process persisted notifications");
                // FIXME : Wait on the queue and not on a flag
                while (NotificationStorageService.this.tryToSendNotification) {
                    try {
                        // FIXME = Do not block forever if the queue is empty,
                        // this will
                        // block the task cancellation and the thread will never
                        // check the tryToSendNotification flag
                        final String id = NotificationStorageService.this.blockingQueue.take();
                        if (NotificationStorageService.this.logger.isLoggable(Level.FINE)) {
                            NotificationStorageService.this.logger
                                    .fine("Processing the file with ID = " + id);
                        }

                        if (POISON.equals(id)) {
                            break;
                        }

                        if (NotificationStorageService.this.getConsumerReference() != null) {

                            // check that the file exists...
                            final File f = new File(NotificationStorageService.this.folder, id);
                            if (f.exists()) {
                                final long now = System.currentTimeMillis();
                                final long fileCreationDate = NotificationStorageService.this
                                        .getFileCreationDate(id);
                                // TODO : the file timestamp can be greater than
                                // the
                                // current time
                                if (TimeUnit.SECONDS.convert(now - fileCreationDate,
                                        TimeUnit.MILLISECONDS) < RESEND_DELAY) {
                                    // TODO : We should change the file
                                    // timestamp so
                                    // that the resend try is exponential
                                    NotificationStorageService.this.blockingQueue.add(id);
                                } else {
                                    // only process the notification which is
                                    // older
                                    // than
                                    // a defined value
                                    try {
                                        final Document doc = XMLUtil
                                                .loadDocument(new FileInputStream(f));
                                        final Notify notify = WSNotificationReader.getInstance()
                                                .readNotify(doc);
                                        NotificationStorageService.this.sender.send(notify,
                                                NotificationStorageService.this
                                                        .getConsumerReference());
                                        // if we are here this is a success, so
                                        // delete the file...
                                        final boolean deleted = f.delete();
                                        if (!deleted) {
                                            NotificationStorageService.this.moveToTrash(f);
                                        }
                                    } catch (final Exception e) {
                                        // change the ID to process the file
                                        // later...
                                        final String newId = NotificationStorageService.this
                                                .getFileName();
                                        final File persistFile = new File(
                                                NotificationStorageService.this.folder,
                                                NotificationStorageService.this.getFileName());
                                        f.renameTo(persistFile);
                                        NotificationStorageService.this.blockingQueue.add(newId);
                                    }
                                }
                            } else {
                                if (NotificationStorageService.this.logger.isLoggable(Level.FINE)) {
                                    NotificationStorageService.this.logger
                                            .fine("File '" + f.getAbsolutePath()
                                                    + "' dos not exists for ID " + id);
                                }
                            }
                        } else {
                            // put the ID for future processing, we are not
                            // registered for now...
                            blockingQueue.add(id);
                        }
                    } catch (final InterruptedException e) {
                        e.printStackTrace();
                    } finally {
                        // ?
                    }
                }
            }
        });
    }

    /**
     * @return
     */
    protected EndpointReferenceType getConsumerReference() {
        return notificationContext.getConsumerReference();
    }

    /**
     * 
     */
    public void stop() {

        // stop the executor
        // TODO : Put a poison pill in the queue to effectively stop the worker
        // (ie not stopped on the take...)
        this.tryToSendNotification = false;
        // if the worker is waiting on the queue.take then add the 'poison pill'
        // to unlock loop...
        this.blockingQueue.add(POISON);
        this.worker.shutdownNow();
        // clear the persisted files...

        this.state = STATE.STOPPED;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void onRegistered(RegisterPublisherResponse response) throws PEtALSCDKException {
        this.processNotifications();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void onUnregistered(DestroyRegistrationResponse response) {

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void onSubscribed(SubscribeResponse response) throws PEtALSCDKException {

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void onUnsubscribed(UnsubscribeResponse response) throws PEtALSCDKException {

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void onException(Exception e) {

    }
}
