/**
 * PETALS: PETALS Services Platform Copyright (C) 2009 EBM WebSourcing
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 * 
 * Initial developer(s): EBM WebSourcing
 */

package org.ow2.petals.component.framework.notification;

import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.component.ComponentContext;
import javax.jbi.messaging.MessagingException;

import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.notification.LifeCyle;
import org.ow2.petals.component.framework.util.RetryTaskExecutor;

import com.ebmwebsourcing.wsstar.addressing.definition.api.EndpointReferenceType;
import com.ebmwebsourcing.wsstar.addressing.definition.api.WSAddressingException;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.Notify;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.Subscribe;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.SubscribeResponse;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.Unsubscribe;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.UnsubscribeResponse;
import com.ebmwebsourcing.wsstar.notification.definition.brokerednotification.api.DestroyRegistration;
import com.ebmwebsourcing.wsstar.notification.definition.brokerednotification.api.DestroyRegistrationResponse;
import com.ebmwebsourcing.wsstar.notification.definition.brokerednotification.api.RegisterPublisher;
import com.ebmwebsourcing.wsstar.notification.definition.brokerednotification.api.RegisterPublisherResponse;
import com.ebmwebsourcing.wsstar.notification.definition.utils.WSNotificationException;
import com.ebmwebsourcing.wsstar.notification.extension.utils.WSNotificationExtensionException;

/**
 * A sender implementation overriding simple action from base sender
 * implementation in order to delegate or enrich work (ie persist things, launch
 * in separate threads, etc...)
 * 
 * @author chamerling - PetalsLink
 * 
 */
public class WrappedSender extends Sender implements
        org.ow2.petals.component.framework.api.notification.Sender, LifeCyle {

    private Logger logger;

    private final NotificationStorageService storageService;

    private final NotificationContext notificationContext;

    private ExecutorService registrationExecutor;

    private ExecutorService executor;

    private NotificationService notificationService;

    /**
     * @param componentContext
     * @throws WSNotificationException
     * @throws MessagingException
     */
    public WrappedSender(final ComponentContext componentContext,
            NotificationService notificationService,
            final NotificationStorageService storageService,
            final NotificationContext notificationContext, Logger logger)
            throws WSNotificationException, MessagingException {
        super(componentContext);
        this.notificationService = notificationService;
        this.storageService = storageService;
        this.notificationContext = notificationContext;
        this.logger = logger;
    }

    /**
     * Wrap the {@inheritDoc}
     */
    @Override
    public void send(final Notify notify, final EndpointReferenceType consumerReference)
            throws WSAddressingException, WSNotificationExtensionException,
            WSNotificationException, MessagingException, PEtALSCDKException {
        // TODO : Must persist the notification somewhere if the subscription is
        // not effective...
        if (!this.isRegistered()) {
            if (this.logger.isLoggable(Level.FINE)) {
                this.logger.log(Level.FINE,
                        "Not registered as publisher, persist the notification for future send");
            }
            // persist the notification if not registered, the persistance
            // service will resend all when available.
            this.persist(notify);
        } else {
            boolean failure = false;
            try {
                if (this.logger.isLoggable(Level.FINE)) {
                    this.logger.log(Level.FINE, "Normal notification sending...");
                }
                super.send(notify, consumerReference);
            } catch (final MessagingException e) {
                if (this.logger.isLoggable(Level.FINE)) {
                    this.logger.log(Level.WARNING, "Can not send the notification", e);
                } else {
                    this.logger.fine("Can not send the notification");
                }
                failure = true;
            } catch (final PEtALSCDKException e) {
                if (this.logger.isLoggable(Level.FINE)) {
                    this.logger.log(Level.WARNING, "Can not send the notification", e);
                } else {
                    this.logger.fine("Can not send the notification");
                }
                failure = true;
            } finally {
                if (failure) {
                    if (this.logger.isLoggable(Level.FINE)) {
                        this.logger
                                .log(Level.FINE,
                                        "A problem occured during notification emission, persist the notification for future emission");
                    }
                    // persist for future send only if the problem is due to the
                    // bus
                    this.persist(notify);
                }
            }
        }
    }

    /**
     * @return
     */
    private boolean isRegistered() {
        return this.notificationContext.isRegistered();
    }

    /**
     * Delegate the send to an executor which wil try to subscribe N times.
     * {@inheritDoc}
     */
    @Override
    public RegisterPublisherResponse send(final RegisterPublisher registerPublisher)
            throws MessagingException, WSNotificationException, PEtALSCDKException {
        RegisterPublisherResponse response = null;
        if (this.registrationExecutor != null) {
            throw new PEtALSCDKException(
                    "Can not launch more than 1 request to register as publisher");
        }

        this.registrationExecutor = Executors.newSingleThreadExecutor();
        final RegistrationTask registrationTask = new RegistrationTask(registerPublisher);
        Future<RegisterPublisherResponse> f = this.registrationExecutor.submit(registrationTask);
        try {
            response = f.get();
        } catch (InterruptedException e) {
            e.printStackTrace();
        } catch (ExecutionException e) {
            e.printStackTrace();
        }
        if (response != null) {
            this.notificationService.onRegistered(response);
        }
        return response;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public DestroyRegistrationResponse send(DestroyRegistration destroyRegistration)
            throws MessagingException, WSNotificationException, PEtALSCDKException {
        DestroyRegistrationResponse result = super.send(destroyRegistration);
        if (result != null) {
            this.notificationService.onUnregistered(result);
        }
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public SubscribeResponse send(Subscribe subscribe) throws MessagingException,
            WSNotificationException, PEtALSCDKException {
        SubscribeResponse response = null;
        final SubscribeTask subscribeTask = new SubscribeTask(subscribe);
        Future<SubscribeResponse> f = this.executor.submit(subscribeTask);
        try {
            response = f.get();
        } catch (InterruptedException e) {
            e.printStackTrace();
        } catch (ExecutionException e) {
            e.printStackTrace();
        }
        if (response != null) {
            this.notificationService.onSubscribed(response);
        }
        return response;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public UnsubscribeResponse send(Unsubscribe unsubscribe) throws MessagingException,
            WSNotificationException, PEtALSCDKException {
        UnsubscribeResponse result = super.send(unsubscribe);
        if (result != null) {
            this.notificationService.onUnsubscribed(result);
        }
        return result;
    }

    /**
     * @param notify
     */
    private void persist(final Notify notify) {
        try {
            this.storageService.store(notify);
        } catch (final PEtALSCDKException e) {
            e.printStackTrace();
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void init() throws PEtALSCDKException {
        this.executor = Executors.newCachedThreadPool();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void start() throws PEtALSCDKException {

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void stop() throws PEtALSCDKException {
        if (this.registrationExecutor != null) {
            this.registrationExecutor.shutdownNow();
        }
        this.registrationExecutor = null;

        if (this.executor != null) {
            this.executor.shutdownNow();
        }
        this.executor = null;
    }

    /**
     * A thread which will try to register the component to the broker...
     * 
     * @author chamerling - eBM WebSourcing
     * 
     */
    class RegistrationTask implements Callable<RegisterPublisherResponse> {

        private RetryTaskExecutor executor;

        private final RegisterPublisher registerPublisher;

        /**
         * 
         */
        public RegistrationTask(final RegisterPublisher registerPublisher) {
            this.registerPublisher = registerPublisher;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public RegisterPublisherResponse call() throws Exception {
            RegisterPublisherResponse result = null;
            this.executor = new RetryTaskExecutor(WrappedSender.this.notificationContext
                    .getMinDelay(), WrappedSender.this.notificationContext.getMaxDelay(),
                    TimeUnit.SECONDS, WrappedSender.this.notificationContext.getDelayFactor(),
                    WrappedSender.this.notificationContext.getNbRetry());
            final Callable<RegisterPublisherResponse> callable = new Callable<RegisterPublisherResponse>() {
                /**
                 * {@inheritDoc}
                 */
                @Override
                public RegisterPublisherResponse call() throws Exception {
                    if (WrappedSender.this.logger.isLoggable(Level.FINE)) {
                        WrappedSender.this.logger
                                .fine("Trying to register myself as publisher to the broker...");
                    }
                    return WrappedSender.super.send(RegistrationTask.this.registerPublisher);
                }
            };
            try {
                result = this.executor.tryExecute(callable);
                // WrappedSender.this.processResponse(response);
            } catch (final Exception e) {
                // ignored (interrupted exception)
                WrappedSender.this.logger.fine(e.getMessage());
            }
            return result;
        }

        /**
         * Stop the registration task
         */
        public void cancel() {
            // stop the executor...
            this.executor.cancel();
        }
    }

    /**
     * Asynchronous send the subscribe payload task N times until someone is
     * able to process it...
     * 
     * @author chamerling - eBM WebSourcing
     * 
     */
    class SubscribeTask implements Callable<SubscribeResponse> {

        private Subscribe subscribe;

        private RetryTaskExecutor executor;

        /**
         * 
         */
        public SubscribeTask(Subscribe subscribe) {
            this.subscribe = subscribe;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public SubscribeResponse call() throws Exception {
            SubscribeResponse result = null;
            this.executor = new RetryTaskExecutor(WrappedSender.this.notificationContext
                    .getMinDelay(), WrappedSender.this.notificationContext.getMaxDelay(),
                    TimeUnit.SECONDS, WrappedSender.this.notificationContext.getDelayFactor(),
                    WrappedSender.this.notificationContext.getNbRetry());
            final Callable<SubscribeResponse> callable = new Callable<SubscribeResponse>() {
                /**
                 * {@inheritDoc}
                 */
                @Override
                public SubscribeResponse call() throws Exception {
                    if (WrappedSender.this.logger.isLoggable(Level.FINE)) {
                        WrappedSender.this.logger
                                .fine("Trying to register myself to the broker as notification receiver...");
                    }
                    return WrappedSender.super.send(SubscribeTask.this.subscribe);
                }
            };
            try {
                result = this.executor.tryExecute(callable);
            } catch (final Exception e) {
                WrappedSender.this.logger.fine(e.getMessage());
            }
            return result;
        }

    }
}
